from typing import List

import fnc

from jesse.config import config
from jesse.models import Order
from jesse.services import selectors
import jesse.helpers as jh


class OrdersState:
    def __init__(self) -> None:
        # used in simulation only
        self.to_execute = []

        self.storage = {}
        self.active_storage = {}

        for exchange in config['app']['trading_exchanges']:
            for symbol in config['app']['trading_symbols']:
                key = f'{exchange}-{symbol}'
                self.storage[key] = []
                self.active_storage[key] = []

    def reset(self) -> None:
        """
        used for testing
        """
        for key in self.storage:
            self.storage[key].clear()
            self.active_storage[key].clear()

    def reset_trade_orders(self, exchange: str, symbol: str) -> None:
        """
        used after each completed trade
        """
        key = f'{exchange}-{symbol}'
        self.storage[key] = []
        self.active_storage[key] = []

    def add_order(self, order: Order) -> None:
        key = f'{order.exchange}-{order.symbol}'
        self.storage[key].append(order)
        self.active_storage[key].append(order)

    def remove_order(self, order: Order) -> None:
        key = f'{order.exchange}-{order.symbol}'
        self.storage[key] = [
            o for o in self.storage[key] if o.id != order.id
        ]
        self.active_storage[key] = [
            o for o in self.active_storage[key] if o.id != order.id
        ]

    def execute_pending_market_orders(self) -> None:
        if not self.to_execute:
            return

        for o in self.to_execute:
            o.execute()

        self.to_execute = []

    # # # # # # # # # # # # # # # # #
    # getters
    # # # # # # # # # # # # # # # # #
    def get_orders(self, exchange, symbol) -> List[Order]:
        key = f'{exchange}-{symbol}'
        return self.storage.get(key, [])

    def get_active_orders(self, exchange, symbol) -> List[Order]:
        key = f'{exchange}-{symbol}'
        return self.active_storage.get(key, [])

    def get_all_orders(self, exchange: str) -> List[Order]:
        return [
            o
            for key in self.storage
            for o in self.storage[key]
            if o.exchange == exchange
        ]

    def count_all_active_orders(self) -> int:
        c = 0
        for key in self.active_storage:
            if len(self.active_storage[key]) == 0:
                continue

            for o in self.active_storage[key]:
                if o.is_active:
                    c += 1
        return c

    def count_active_orders(self, exchange: str, symbol: str) -> int:
        orders = self.get_active_orders(exchange, symbol)

        return sum(bool(o.is_active) for o in orders)

    def count(self, exchange: str, symbol: str) -> int:
        return len(self.get_orders(exchange, symbol))

    def get_order_by_id(self, exchange: str, symbol: str, id: str, use_exchange_id: bool = False) -> Order:
        key = f'{exchange}-{symbol}'

        if use_exchange_id:
            return fnc.find(lambda o: o.exchange_id == id, self.storage[key])

        return fnc.find(lambda o: id in o.id, reversed(self.storage[key]))

    def get_entry_orders(self, exchange: str, symbol: str) -> List[Order]:
        # return all orders if position is not opened yet
        p = selectors.get_position(exchange, symbol)
        if p.is_close:
            return self.get_orders(exchange, symbol).copy()

        all_orders = self.get_active_orders(exchange, symbol)
        p_side = jh.type_to_side(p.type)
        entry_orders = [o for o in all_orders if (o.side == p_side and not o.is_canceled)]

        return entry_orders

    def get_exit_orders(self, exchange: str, symbol: str) -> List[Order]:
        """
        excludes cancel orders but includes executed orders
        """
        all_orders = self.get_orders(exchange, symbol)
        # return empty if no orders
        if len(all_orders) == 0:
            return []
        # return empty if position is not opened yet
        p = selectors.get_position(exchange, symbol)
        if p.is_close:
            return []
        else:
            exit_orders = [o for o in all_orders if o.side != jh.type_to_side(p.type)]

        # exclude cancelled orders
        exit_orders = [o for o in exit_orders if not o.is_canceled]

        return exit_orders

    def get_active_exit_orders(self, exchange: str, symbol: str) -> List[Order]:
        """
        excludes cancel orders but includes executed orders
        """
        all_orders = self.get_active_orders(exchange, symbol)
        # return empty if no orders
        if len(all_orders) == 0:
            return []
        # return empty if position is not opened yet
        p = selectors.get_position(exchange, symbol)
        if p.is_close:
            return []
        else:
            exit_orders = [o for o in all_orders if o.side != jh.type_to_side(p.type)]

        # exclude cancelled orders
        exit_orders = [o for o in exit_orders if not o.is_canceled]

        return exit_orders

    def update_active_orders(self, exchange: str, symbol: str):
        key = f'{exchange}-{symbol}'
        active_orders = [
            order
            for order in self.get_active_orders(exchange, symbol)
            if not order.is_canceled and not order.is_executed
        ]
        self.active_storage[key] = active_orders



from jesse.enums import exchanges as exchanges_enums, timeframes

JESSE_API_URL = 'https://api1.jesse.trade/api'
# JESSE_API_URL = 'http://localhost:8040/api'
JESSE_WEBSITE_URL = 'https://jesse.trade'
# JESSE_WEBSITE_URL = 'http://localhost:8040'

BYBIT_TIMEFRAMES = [timeframes.MINUTE_1, timeframes.MINUTE_3, timeframes.MINUTE_5, timeframes.MINUTE_15, timeframes.MINUTE_30,
                    timeframes.HOUR_1, timeframes.HOUR_2, timeframes.HOUR_4, timeframes.HOUR_6, timeframes.HOUR_12, timeframes.DAY_1]
FTX_TIMEFRAMES = [timeframes.MINUTE_1, timeframes.MINUTE_3, timeframes.MINUTE_5, timeframes.MINUTE_15, timeframes.MINUTE_30,
                  timeframes.HOUR_1, timeframes.HOUR_2, timeframes.HOUR_4, timeframes.HOUR_6, timeframes.HOUR_12, timeframes.DAY_1]
BINANCE_TIMEFRAMES = [timeframes.MINUTE_1, timeframes.MINUTE_3, timeframes.MINUTE_5, timeframes.MINUTE_15, timeframes.MINUTE_30,
                      timeframes.HOUR_1, timeframes.HOUR_2, timeframes.HOUR_4, timeframes.HOUR_6, timeframes.HOUR_8, timeframes.HOUR_12, timeframes.DAY_1]
COINBASE_TIMEFRAMES = [timeframes.MINUTE_1, timeframes.MINUTE_5,
                       timeframes.MINUTE_15, timeframes.HOUR_1, timeframes.HOUR_6, timeframes.DAY_1]
BITGET_TIMEFRAMES = [timeframes.MINUTE_1, timeframes.MINUTE_5, timeframes.MINUTE_15,
                     timeframes.MINUTE_30, timeframes.HOUR_1, timeframes.HOUR_4, timeframes.HOUR_12, timeframes.DAY_1]
DYDX_TIMEFRAMES = [timeframes.MINUTE_1, timeframes.MINUTE_5, timeframes.MINUTE_15,
                   timeframes.MINUTE_30, timeframes.HOUR_1, timeframes.HOUR_4, timeframes.DAY_1]
APEX_PRO_TIMEFRAMES = [timeframes.MINUTE_1, timeframes.MINUTE_5, timeframes.MINUTE_15,
                       timeframes.MINUTE_30, timeframes.HOUR_1, timeframes.HOUR_4, timeframes.DAY_1]

exchange_info = {
    # BYBIT_USDT_PERPETUAL
    exchanges_enums.BYBIT_USDT_PERPETUAL: {
        'name': exchanges_enums.BYBIT_USDT_PERPETUAL,
        'url': JESSE_WEBSITE_URL + '/bybit',
        'fee': 0.00055,
        'type': 'futures',
        'settlement_currency': 'USDT',
        'supported_leverage_modes': ['cross', 'isolated'],
        'supported_timeframes': BYBIT_TIMEFRAMES,
        'modes': {
            'backtesting': True,
            'live_trading': True,
        },
        'required_live_plan': 'premium'
    },
    # BYBIT_USDT_PERPETUAL_TESTNET
    exchanges_enums.BYBIT_USDT_PERPETUAL_TESTNET: {
        'name': exchanges_enums.BYBIT_USDT_PERPETUAL_TESTNET,
        'url': JESSE_WEBSITE_URL + '/bybit',
        'fee': 0.00055,
        'type': 'futures',
        'settlement_currency': 'USDT',
        'supported_leverage_modes': ['cross', 'isolated'],
        'supported_timeframes': BYBIT_TIMEFRAMES,
        'modes': {
            'backtesting': True,
            'live_trading': True,
        },
        'required_live_plan': 'premium'
    },
    # BYBIT_USDT_PERPETUAL
    exchanges_enums.BYBIT_USDC_PERPETUAL: {
        'name': exchanges_enums.BYBIT_USDC_PERPETUAL,
        'url': JESSE_WEBSITE_URL + '/bybit',
        'fee': 0.00055,
        'type': 'futures',
        'settlement_currency': 'USDC',
        'supported_leverage_modes': ['cross', 'isolated'],
        'supported_timeframes': BYBIT_TIMEFRAMES,
        'modes': {
            'backtesting': True,
            'live_trading': True,
        },
        'required_live_plan': 'premium'
    },
    # BYBIT_USDC_PERPETUAL_TESTNET
    exchanges_enums.BYBIT_USDC_PERPETUAL_TESTNET: {
        'name': exchanges_enums.BYBIT_USDC_PERPETUAL_TESTNET,
        'url': JESSE_WEBSITE_URL + '/bybit',
        'fee': 0.00055,
        'type': 'futures',
        'supported_leverage_modes': ['cross', 'isolated'],
        'supported_timeframes': BYBIT_TIMEFRAMES,
        'settlement_currency': 'USDC',
        'modes': {
            'backtesting': True,
            'live_trading': True,
        },
        'required_live_plan': 'premium'
    },
    # BYBIT_SPOT_TESTNET
    exchanges_enums.BYBIT_SPOT: {
        'name': exchanges_enums.BYBIT_SPOT,
        'url': 'https://jesse.trade/bybit',
        'fee': 0.001,
        'type': 'spot',
        'supported_leverage_modes': ['cross', 'isolated'],
        'supported_timeframes': BYBIT_TIMEFRAMES,
        'modes': {
            'backtesting': True,
            'live_trading': True,
        },
        'required_live_plan': 'premium'
    },
    # BYBIT_SPOT_TESTNET
    exchanges_enums.BYBIT_SPOT_TESTNET: {
        'name': exchanges_enums.BYBIT_SPOT_TESTNET,
        'url': 'https://jesse.trade/bybit',
        'fee': 0.001,
        'type': 'spot',
        'supported_leverage_modes': ['cross', 'isolated'],
        'supported_timeframes': BYBIT_TIMEFRAMES,
        'modes': {
            'backtesting': True,
            'live_trading': True,
        },
        'required_live_plan': 'premium'
    },
    # FTX_PERPETUAL_FUTURES
    exchanges_enums.FTX_PERPETUAL_FUTURES: {
        'name': exchanges_enums.FTX_PERPETUAL_FUTURES,
        'url': 'https://ftx.com/markets/future',
        'fee': 0.0006,
        'type': 'futures',
        'supported_leverage_modes': ['cross'],
        'supported_timeframes': FTX_TIMEFRAMES,
        'modes': {
            'backtesting': False,
            'live_trading': False,
        },
        'required_live_plan': 'premium'
    },
    # FTX_SPOT
    exchanges_enums.FTX_SPOT: {
        'name': exchanges_enums.FTX_SPOT,
        'url': 'https://ftx.com/markets/spot',
        'fee': 0.0007,
        'type': 'spot',
        'supported_leverage_modes': ['cross'],
        'supported_timeframes': FTX_TIMEFRAMES,
        'modes': {
            'backtesting': False,
            'live_trading': False,
        },
        'required_live_plan': 'premium'
    },
    # FTX_US_SPOT
    exchanges_enums.FTX_US_SPOT: {
        'name': exchanges_enums.FTX_US_SPOT,
        'url': 'https://ftx.us',
        'fee': 0.002,
        'type': 'spot',
        'supported_leverage_modes': ['cross'],
        'supported_timeframes': FTX_TIMEFRAMES,
        'modes': {
            'backtesting': False,
            'live_trading': False,
        },
        'required_live_plan': 'premium'
    },
    # BITFINEX_SPOT
    exchanges_enums.BITFINEX_SPOT: {
        'name': exchanges_enums.BITFINEX_SPOT,
        'url': 'https://bitfinex.com',
        'fee': 0.002,
        'type': 'spot',
        'supported_leverage_modes': ['cross'],
        'supported_timeframes': [timeframes.MINUTE_1, timeframes.MINUTE_5, timeframes.MINUTE_15, timeframes.MINUTE_30, timeframes.HOUR_1, timeframes.HOUR_3, timeframes.HOUR_6, timeframes.HOUR_12, timeframes.DAY_1],
        'modes': {
            'backtesting': True,
            'live_trading': False,
        },
        'required_live_plan': 'premium'
    },
    # BINANCE_SPOT
    exchanges_enums.BINANCE_SPOT: {
        'name': exchanges_enums.BINANCE_SPOT,
        'url': 'https://binance.com',
        'fee': 0.001,
        'type': 'spot',
        'supported_leverage_modes': ['cross', 'isolated'],
        'supported_timeframes': BINANCE_TIMEFRAMES,
        'modes': {
            'backtesting': True,
            'live_trading': True,
        },
        'required_live_plan': 'premium'
    },
    # BINANCE_US_SPOT
    exchanges_enums.BINANCE_US_SPOT: {
        'name': exchanges_enums.BINANCE_US_SPOT,
        'url': 'https://binance.us',
        'fee': 0.001,
        'type': 'spot',
        'supported_leverage_modes': ['cross', 'isolated'],
        'supported_timeframes': BINANCE_TIMEFRAMES,
        'modes': {
            'backtesting': True,
            'live_trading': True,
        },
        'required_live_plan': 'premium'
    },
    # BINANCE_PERPETUAL_FUTURES
    exchanges_enums.BINANCE_PERPETUAL_FUTURES: {
        'name': exchanges_enums.BINANCE_PERPETUAL_FUTURES,
        'url': 'https://binance.com',
        'fee': 0.0004,
        'type': 'futures',
        'supported_leverage_modes': ['cross', 'isolated'],
        'supported_timeframes': BINANCE_TIMEFRAMES,
        'modes': {
            'backtesting': True,
            'live_trading': True,
        },
        'required_live_plan': 'premium'
    },
    # BINANCE_PERPETUAL_FUTURES_TESTNET
    exchanges_enums.BINANCE_PERPETUAL_FUTURES_TESTNET: {
        'name': exchanges_enums.BINANCE_PERPETUAL_FUTURES_TESTNET,
        'url': 'https://binance.com',
        'fee': 0.0004,
        'type': 'futures',
        'supported_leverage_modes': ['cross', 'isolated'],
        'supported_timeframes': BINANCE_TIMEFRAMES,
        'modes': {
            'backtesting': True,
            'live_trading': True,
        },
        'required_live_plan': 'premium'
    },
    # COINBASE_SPOT
    exchanges_enums.COINBASE_SPOT: {
        'name': exchanges_enums.COINBASE_SPOT,
        'url': 'https://www.coinbase.com/advanced-trade/spot/BTC-USD',
        'fee': 0.0003,
        'type': 'spot',
        'supported_leverage_modes': ['cross', 'isolated'],
        'supported_timeframes': COINBASE_TIMEFRAMES,
        'modes': {
            'backtesting': True,
            'live_trading': True,
        },
        'required_live_plan': 'premium'
    },
    # BITGET_USDT_PERPETUAL_TESTNET
    exchanges_enums.BITGET_USDT_PERPETUAL_TESTNET: {
        'name': exchanges_enums.BITGET_USDT_PERPETUAL_TESTNET,
        'url': JESSE_WEBSITE_URL + '/bitget',
        'fee': 0.0006,
        'type': 'futures',
        'supported_leverage_modes': ['cross', 'isolated'],
        'supported_timeframes': BITGET_TIMEFRAMES,
        'modes': {
            'backtesting': False,
            'live_trading': False,
        },
        'required_live_plan': 'premium'
    },
    # BITGET_USDT_PERPETUAL
    exchanges_enums.BITGET_USDT_PERPETUAL: {
        'name': exchanges_enums.BITGET_USDT_PERPETUAL,
        'url': JESSE_WEBSITE_URL + '/bitget',
        'fee': 0.0006,
        'type': 'futures',
        'supported_leverage_modes': ['cross', 'isolated'],
        'supported_timeframes': BITGET_TIMEFRAMES,
        'modes': {
            'backtesting': False,
            'live_trading': True,
        },
        'required_live_plan': 'premium'
    },
    # BITGET_SPOT
    exchanges_enums.BITGET_SPOT: {
        'name': exchanges_enums.BITGET_SPOT,
        'url': JESSE_WEBSITE_URL + '/bitget',
        'fee': 0.0006,
        'type': 'spot',
        'supported_leverage_modes': ['cross', 'isolated'],
        'supported_timeframes': BITGET_TIMEFRAMES,
        'modes': {
            'backtesting': False,
            # disabled for now
            'live_trading': False,
        },
        'required_live_plan': 'premium'
    },
    # DyDx
    exchanges_enums.DYDX_PERPETUAL: {
        'name': exchanges_enums.DYDX_PERPETUAL,
        'url': JESSE_WEBSITE_URL + '/dydx',
        'fee': 0.0005,
        'type': 'futures',
        'supported_leverage_modes': ['cross'],
        'supported_timeframes': DYDX_TIMEFRAMES,
        'modes': {
            'backtesting': False,
            'live_trading': True,
        },
        'required_live_plan': 'premium'
    },
    # DyDx Testnet
    exchanges_enums.DYDX_PERPETUAL_TESTNET: {
        'name': exchanges_enums.DYDX_PERPETUAL_TESTNET,
        'url': 'https://trade.stage.dydx.exchange/trade/ETH-USD',
        'fee': 0.0005,
        'type': 'futures',
        'supported_leverage_modes': ['cross'],
        'supported_timeframes': DYDX_TIMEFRAMES,
        'modes': {
            'backtesting': False,
            'live_trading': True,
        },
        'required_live_plan': 'premium'
    },

    exchanges_enums.APEX_PRO_PERPETUAL_TESTNET: {
        'name': exchanges_enums.APEX_PRO_PERPETUAL_TESTNET,
        'url': 'https://testnet.pro.apex.exchange/trade/BTCUSD',
        'fee': 0.0005,
        'type': 'futures',
        'supported_leverage_modes': ['cross'],
        'supported_timeframes': APEX_PRO_TIMEFRAMES,
        'modes': {
            'backtesting': False,
            'live_trading': True,
        },
        'required_live_plan': 'premium'
    },

    exchanges_enums.APEX_PRO_PERPETUAL: {
        'name': exchanges_enums.APEX_PRO_PERPETUAL,
        'url': 'https://pro.apex.exchange/trade/BTCUSD',
        'fee': 0.0005,
        'type': 'futures',
        'supported_leverage_modes': ['cross'],
        'supported_timeframes': APEX_PRO_TIMEFRAMES,
        'modes': {
            'backtesting': False,
            'live_trading': True,
        },
        'required_live_plan': 'premium'
    }

}

# list of supported exchanges for backtesting
backtesting_exchanges = [k for k, v in exchange_info.items() if v['modes']['backtesting'] is True]
backtesting_exchanges = list(sorted(backtesting_exchanges))

# list of supported exchanges for live trading
live_trading_exchanges = [k for k, v in exchange_info.items() if v['modes']['live_trading'] is True]
live_trading_exchanges = list(sorted(live_trading_exchanges))

# used for backtesting, and live trading when local candle generation is enabled:
jesse_supported_timeframes = [
    timeframes.MINUTE_1,
    timeframes.MINUTE_3,
    timeframes.MINUTE_5,
    timeframes.MINUTE_15,
    timeframes.MINUTE_30,
    timeframes.MINUTE_45,
    timeframes.HOUR_1,
    timeframes.HOUR_2,
    timeframes.HOUR_3,
    timeframes.HOUR_4,
    timeframes.HOUR_6,
    timeframes.HOUR_8,
    timeframes.HOUR_12,
    timeframes.DAY_1,
]



from .BybitMain import BybitMain
from jesse.enums import exchanges


class BybitSpotTestnet(BybitMain):
    def __init__(self) -> None:
        super().__init__(
            name=exchanges.BYBIT_SPOT_TESTNET,
            rest_endpoint='https://api-testnet.bybit.com',
            category='spot',
        )



from jesse.enums import timeframes


def timeframe_to_interval(timeframe: str) -> str:
    if timeframe == timeframes.MINUTE_1:
        return '1min'
    elif timeframe == timeframes.MINUTE_5:
        return '5min'
    elif timeframe == timeframes.MINUTE_15:
        return '15min'
    elif timeframe == timeframes.MINUTE_30:
        return '30min'
    elif timeframe == timeframes.HOUR_1:
        return '1h'
    elif timeframe == timeframes.HOUR_4:
        return '4h'
    elif timeframe == timeframes.HOUR_6:
        return '6h'
    elif timeframe == timeframes.HOUR_12:
        return '12h'
    elif timeframe == timeframes.DAY_1:
        return '1day'


def interval_to_timeframe(interval: str) -> str:
    if interval == '1min':
        return timeframes.MINUTE_1
    elif interval == '5min':
        return timeframes.MINUTE_5
    elif interval == '15min':
        return timeframes.MINUTE_15
    elif interval == '30min':
        return timeframes.MINUTE_30
    elif interval == '1h':
        return timeframes.HOUR_1
    elif interval == '4h':
        return timeframes.HOUR_4
    elif interval == '6h':
        return timeframes.HOUR_6
    elif interval == '12h':
        return timeframes.HOUR_12
    elif interval == '1day':
        return timeframes.DAY_1



from jesse.enums import timeframes


def timeframe_to_interval(timeframe: str) -> str:
    # 1DAY, 4HOURS, 1HOUR, 30MINS, 15MINS, 5MINS, 1MIN
    if timeframe == timeframes.MINUTE_1:
        return '1MIN'
    elif timeframe == timeframes.MINUTE_5:
        return '5MINS'
    elif timeframe == timeframes.MINUTE_15:
        return '15MINS'
    elif timeframe == timeframes.MINUTE_30:
        return '30MINS'
    elif timeframe == timeframes.HOUR_1:
        return '1HOUR'
    elif timeframe == timeframes.HOUR_4:
        return '4HOURS'
    elif timeframe == timeframes.DAY_1:
        return '1DAY'
    else:
        raise ValueError('Invalid timeframe: {}'.format(timeframe))



from jesse.enums import timeframes
import jesse.helpers as jh


def timeframe_to_interval(timeframe: str) -> str:
    if timeframe == timeframes.MINUTE_1:
        return '1m'
    elif timeframe == timeframes.MINUTE_5:
        return '5m'
    elif timeframe == timeframes.MINUTE_15:
        return '15m'
    elif timeframe == timeframes.MINUTE_30:
        return '30m'
    elif timeframe == timeframes.HOUR_1:
        return '1H'
    elif timeframe == timeframes.HOUR_4:
        return '4H'
    elif timeframe == timeframes.HOUR_12:
        return '12H'
    elif timeframe == timeframes.DAY_1:
        return '1D'
    else:
        raise NotImplemented('Invalid timeframe: {}'.format(timeframe))


def interval_to_timeframe(interval: str) -> str:
    if interval == '1m':
        return timeframes.MINUTE_1
    elif interval == '5m':
        return timeframes.MINUTE_5
    elif interval == '15m':
        return timeframes.MINUTE_15
    elif interval == '30m':
        return timeframes.MINUTE_30
    elif interval == '1H':
        return timeframes.HOUR_1
    elif interval == '4H':
        return timeframes.HOUR_4
    elif interval == '12H':
        return timeframes.HOUR_12
    elif interval == '1D':
        return timeframes.DAY_1
    else:
        raise NotImplemented('Invalid interval: {}'.format(interval))



from .BitgetUSDTPerpetualMain import BitgetUSDTPerpetualMain
from jesse.enums import exchanges


class BitgetUSDTPerpetual(BitgetUSDTPerpetualMain):
    def __init__(self) -> None:
        super().__init__(
            name=exchanges.BITGET_USDT_PERPETUAL,
            endpoint='https://api.bitget.com'
        )
